package GeoDB;

use strict;

use GeoDB::Utils;
use GeoDB::DB;
use GeoDB::Set;
use GeoDB::Waypoint;
use GeoDB::Waypoint::Geocache;
use GeoDB::Waypoint::Waymark;
use GeoDB::Waypoint::Wigle;
use GeoDB::Waypoint::Geodining;

our $VERSION = "0.71";
our @ISA=qw(GeoDB::DB);

our %sqlxlat =
  ( '&&' => ' and ',
    '||' => ' or ',
    ')' => ')',
    '(' => '(',
    '' => '',
  );

# note: this new clause is used by most sub-modules too, altering it
# will alter them.
sub new {
    my $type = shift;
    my ($class) = ref($type) || $type;
    my $self = {};
    %$self = @_;
    bless($self, $class);
    $self->init_global($self->{'dbh'}) if ($self->{'dbh'});
    return $self;
}

######################################################################
# INITIALIZATION
#
sub init {
    my $self = shift;
    return $self->init_global(@_);
}

my $have_done_global_init = 0;
# initial needed setup by various modules
sub init_global {
    my $self = shift;

    return 0 if (!$self->{'dbh'});

    # run only once ever
    return 1 if ($have_done_global_init);
    $have_done_global_init = 1;

    # call modules that need it.

    # sub modules of waypoint MUST be called first to register themselves
    # so that the Waypoint can init based on those registrations.
    GeoDB::Waypoint::Geocache::init($self->{'dbh'});
    GeoDB::Waypoint::Waymark::init($self->{'dbh'});
    GeoDB::Waypoint::Wigle::init($self->{'dbh'});
    GeoDB::Waypoint::Geodining::init($self->{'dbh'});
    GeoDB::Waypoint::init($self->{'dbh'});
    return 1;
}

######################################################################
# ENGINE FUNCTIONS

#
# tries to require (load) and call new() on a generic perl module
#
sub new_engine {
    my ($self, $type, $parents, $superparent) = @_;
    my ($args, $argsep);

    $superparent = "GeoDB" if (!$superparent);
    $parents = "Search" if (!$parents);
    $parents = [$parents, "Everywhere"] if (ref($parents) ne 'ARRAY');

    # some things may pass :args terminated
    ($type, $argsep, $args) = ($type =~ /([\/\w]+)(:*)(.*)$/);

    # replace slashes with ::s
    $type =~ s/\//::/g;

    # capitalize first letter only of each module (enforced convention)
    $type =~ s/^(.)(.*)/uc($1) . lc($2)/e;
    $type =~ s/::(.)([^:]*)/"::" . uc($1) . lc($2)/ge;

    my $engine;
    my $err;
    for (my $i = 0; $i <= $#$parents && !$engine; $i++ ) {
	my $parent = $parents->[$i];
	GEODEBUG(6,
		 "Trying to load module '${superparent}::${parent}::$type'\n");

	# try and see if we can require the module
	# (if yes, it exists; if not skip it)
	my $res = eval "require ${superparent}::${parent}::$type;";
	if (!$res) {
	    my $potentialerr = $@;
	    if (!$err && $potentialerr !~ /Can.t locate/) {
		$err = "\n($potentialerr)";
	    }
	    next;
	}

	$engine = eval "new ${superparent}::${parent}::$type (args => \"$args\", type => \"$parents->[0]\")" ||
	  GEOERR("can't create a $parent engine for type='${superparent}::${parent}::$type';\n   (frequently syntax issue, no 'new' operator, or wrong package name ($@)\n");

	if (ref($engine) ne '') {
	    $engine->set_master($self);
	    if (!$argsep && !$self->get_option('nogui')) {
		$engine->prompt_args();
	    } else {
		$engine->parse_args($args);
	    }
	}

	return $engine;
    }
    GEOERR("Could find $parents->[0] Engine Type '$type' $err\n");
}

sub create_question_from_args {
    my ($self, $opt, $default) = @_;
    next if ($opt->{'nogui'});
    my $q =
      { type => 'text',
	name => $opt->{'name'},
	text => $opt->{'title'} || "$opt->{'name'}",
	helptext => $opt->{'help'},
	size => 60,
      };
    $q->{'default'} = $default if ($default);
    foreach my $passon (qw(type values labels indent helpdesc text doif
			   refresh_on_change default submit)) {
	if (exists($opt->{$passon})) {
	    $q->{$passon} = $opt->{$passon};
	    if ($q->{$passon} =~ /engine:(.*)/) {
		my $e = $self->new_engine($1, ["AutoGUI::Values"]);
		$q->{$passon} = $e->get_gui_values($q->{$passon});
	    }
	}
    }

    # special processing for most of the above (sigh)

    # values/labels become an array
    $q->{'values'} = [split(/,/,$q->{'values'})]
      if(exists($q->{'values'}) && ref($q->{'values'}) eq '');
    $q->{'labels'} = [split(/,/,$q->{'labels'})]
      if(exists($q->{'labels'})  && ref($q->{'values'}) eq '');

    # doif is either a single word, or a word=value
    # (XXX: eventually allow arbitrary code refs somehow?)
    $q->{'doif'} = eval "sub { return qwparam($q->{'doif'}); }"
      if (exists($q->{'doif'}) && ref($q->{'doif'}) ne 'CODE' && 
	  $q->{'doif'} !~ /=/);
    if (exists($q->{'doif'}) && ref($q->{'doif'}) ne 'CODE' && 
	$q->{'doif'} =~ /=/) {
	my ($left, $right) = ($q->{'doif'} =~ /([^=]+)=(.*)/);
	$q->{'doif'} =
	  eval "sub { return (qwparam(\"$left\") eq \"$right\"); }";
    }

    # Auto-set the type to 'menu' if there is a labels or a value clause
    $q->{'type'} = 'menu'
      if ((exists($q->{'labels'}) || exists($q->{'values'})) &&
	  !exists($opt->{'type'}));
    return $q;
}

sub prompt_args {
    my ($self) = @_;
    my $qw = $self->get_qwizard();
    return if (!$qw);

    my $pluginname = ref($self);
    my $displayname = $pluginname;
    $displayname =~ s/^GeoDB:://;
    $displayname =~ s/::/ /;
    $displayname =~ s/::/\//g;
    my $options = eval "[\@${pluginname}::GEOOPTIONS];";
    my @options =@$options if ($options);

    return if ($#options == -1);

    my @qs;
    my $q;
    foreach my $opt (@options) {
	next if ($opt->{'nogui'});
	$q = $self->create_question_from_args($opt);
	push @qs, $q;
    }

    $qw->merge_primaries(
    {
     'askqs' => 
     { title => "Parameters for the $displayname plugin",
       questions => \@qs,
       post_answers => 
       [[sub {
	    my $opts = $_[$#_-1];
	    my $self = $_[$#_];
	    foreach my $opt (@options) {
		$self->{'options'}{$opt->{'name'}} = qwparam($opt->{'name'})
		  if (qwparam($opt->{'name'}) ne '');
	    }
	},\@options, $self]],
     },
    }
			);
    $qw->magic('askqs');
}

sub set_master {
    my ($self, $master) = @_;
    $self->{'master'} = $master;
}

sub get_master {
    my ($self) = @_;
    while (exists($self->{'master'})) {
	$self = $self->{'master'};
    }
    return $self;
}

sub get_option {
    my ($self, $opt) = @_;
    $self = $self->get_master();
    return $self->{'main_options'}{$opt} if (exists($self->{'main_options'}));
    return $main::opts{$opt};
}

sub set_option {
    my ($self, $opt, $value) = @_;
    $self = $self->get_master();
    if (exists($self->{'main_options'})) {
	$self->{'main_options'}{$opt} = $value;
    } else {
	$main::opts{$opt} = $value;
    }
}

sub save_option {
    my ($self, $opt, $value) = @_;
    $self = $self->get_master();
    return if ($self->get_option($opt) eq $value);
    $self->set_option($opt, $value);
    GEODEBUG(5, "saving option $opt, to $self->{'default_file'}\n");
    if ($self->{'default_file'}) {
	open(O,">>$self->{'default_file'}");
	print O "$opt $value\n";
	close(O);
    }
}


sub get_default {
    my ($self, $opt) = @_;
    $self = $self->get_master();
    return $self->{'main_options'}{'defaults'}{$opt}
      if (exists($self->{'main_options'}) &&
	  exists($self->{'main_options'}{'defaults'}));
    return $main::opts{'defaults'}{$opt};
}

sub save_default {
    my ($self, $opt, $value) = @_;
    $self = $self->get_master();
    return if ($self->get_default($opt) eq $value);
    $self->{'main_options'}{'defaults'}{$opt} = $value;
    if ($self->{'default_file'}) {
	open(O,">>$self->{'default_file'}");
	print O "default $opt $value\n";
	close(O);
    }
}

#
# parses specs of the form:
#   engtype1:arg1=arg2+engtype2:arg3=arg4:out+...
# returns an array of engines
#
sub get_engines_from_specs {
    my ($self, $specs, $parent, $superparent) = @_;
    my $args;
    my @engines;
    $specs = [$specs] if (ref($specs) ne 'ARRAY');

    foreach my $spec (@$specs) {
	while ($spec =~ s/^([^\+]+)(\+?)//) {
	    my $part = $1;
	    my $sep = $2;
	    GEODEBUG(3,"processing engine spec: $part => $spec\n");
	    my $e = $self->new_engine($part, $parent, $superparent);
	    if (ref($e) eq '') {
		$spec = $e . $sep . $spec;
	    } else {
		push @engines, $e;
	    }
	}
    }
    return \@engines;
}

#
# constructs a search "where" clause and arguments from a search spec string
#
sub construct_search {
    my ($self, $arg) = @_;

    my ($whereclause, $newwhereclause, $newargs);
    my @args;

    while ($arg) {
	if ($arg =~ s/^(\(|\))//) {
	    # allow (, or ) to pass through
	    $whereclause .= $1;
	} elsif ($arg =~ s/^(\&\&|\|\|)//) {
	    # allow && or || to pass through translated
	    $whereclause .= $sqlxlat{$1};
	} elsif ($arg =~ s/(\w+:|)([^\&\|\(\)]*)(\&\&|\|\||)//) {
	    # type:args and optionally &&, ||, ( or )
	    my ($type, $term, $clause) = ($1, $2, $3);

	    GEODEBUG(2,"Parsing search term: $type / $term / $clause\n");

	    GEODEBUG(5," type: $type\n");
	    my $engine = $self->new_engine("$type:$term");
	    if (ref($engine) eq '') {
		$arg = $engine . $arg;
		next;
	    }
	    ($newwhereclause, $newargs) = $engine->search($term, $type);
	    $whereclause .= " ($newwhereclause) $sqlxlat{$clause}";
	    GEODEBUG(3,"where clause part:      $newwhereclause\n");
	    GEODEBUG(3,"where clause arg count: $#$newargs\n");
	    push @args, @$newargs;
	} else {
	    print STDERR ("CAN'T PARSE ARGS:" . $arg);
	    exit;
	}
    }
    return ($whereclause, \@args);
}

# note: import is a reserved function name in perl, so we had to use import_it
sub import_it {
    my ($self, @files) = @_;
    my $set;
    foreach my $f (@files) {
	my $parenttype = "Import";
	GEODEBUG(1,"importing $f\n");
	while ($f) {
	    $f =~ s/([^\+;]+)([\+;]*)(.*)/$3/;
	    my ($it, $sep) = ($1, $2);
	    if (!$it) {
		$it = $f;
		$f = '';
	    }
	    if ($parenttype eq "Import") {
		my ($type, $file) = get_type_and_file($it);
		my $engine = $self->new_engine($type, $parenttype);
		if (ref($engine) !~ /GeoDB/) {
		    # must be a string replacement
		    $f = $engine . $sep . $f;
		    next;
		}
		$set->save($self->{'dbh'}) if ($set);
	        $set = $engine->import_it($file, $self->{'dbh'});
	    } else {
		my $engine = $self->new_engine($it, $parenttype);
		if (ref($engine) !~ /GeoDB/) {
		    # must be a string replacement
		    $f = $engine . $sep . $f;
		    next;
		}
		$engine->modify_set($set);
	    }
	    $parenttype = ((!$sep || $sep eq ';') ? "Import" : "Modify");
	}
	$set->save($self->{'dbh'}) if ($set);
	$set = undef;
    }
    return $set;
}

sub modify {
    my ($self, $set, $modifydescs) = @_;

    GEODEBUG(2,"modifying set as requested\n");

    my $engines = $self->get_engines_from_specs($modifydescs, "Modify");

    foreach my $engine (@$engines) {
	$engine->modify_set($set);
    }
}

#
# export data
#
sub export_it {
    my ($self, $exportdescs, $origset) = @_;
    $exportdescs = [$exportdescs] if (ref($exportdescs) ne 'ARRAY');

    my $set = $origset->clone();

    foreach my $f (@$exportdescs) {
	# modtype1:args+modtype2:args+outputfile.type
	GEODEBUG(5,"export spec: $f\n");
	while ($f) {
	    GEODEBUG(6,"export processing: $f\n");
	    # contains a modify part
	    my $sep;
	    ($sep) = ($f =~ /([\+;])/);
	    if ($sep && $sep eq '+') {
		$f =~ s/([^\+;]+)\+//;
		my $part = $1;
		my $engine = $self->new_engine($part, "Modify");
		if (ref($engine) eq '') {
		    # it returned a string; keep going with it
		    $f = $engine;
		} else {
		    # no other parts left, this is indeed the export engine
		    GEODEBUG(2,"modifying using " . ref($engine) . "\n");
		    $engine->modify_set($set);
		}

   	    # contains a single part, potentially the last export part
	    } else {
		$f =~ s/([^;]+);*//;
		my $part = $1;

		my ($type, $file);
		if ($part !~ /:/) {
		    ($type, $file) = get_type_and_file($part);
		    $part = "$type:$file";
		}
		my $engine = $self->new_engine($part, "Export");
		if (ref($engine) eq '') {
		    # it returned a string; keep going with it
		    $f = $engine . ($sep || "") . $f;
		} else {
		    # no other parts left, this is indeed the export engine
		    GEODEBUG(2,"exporting using " . ref($engine). "\n");
		    # pull out the args differently than the normal parsing
		    ($type, $file) = get_type_and_file($part) if (!$file);
		    my $engine = $self->new_engine($type, "Export");
		    $engine->export_it($file, $set);
		    $set = $origset->clone() if ($f);
		}
	    }
	}
    }
}

#
# take a search term (arg) and load it into (set or created) using options.
#   - returns either the passed set or a new one if one wasn't passed.
sub search {
    my ($self, $arg, $options) = @_;

    my $searchset = $options->{'searchset'} || GeoDB::Set->new();

    my ($whereclause, $args) = $self->construct_search($arg);

    GEODEBUG(2,"final search clause: $whereclause\n");
    GEODEBUG(2,"final search args:   ", join(", ",@$args),"\n");

    my $ids = get_caches_from_where_clause($self->{'dbh'}, $whereclause, $args,
					   $options->{'limit'});
    if ($ids) {
	$searchset->add_by_ids(@$ids);
    } else {
	GEOERR("no waypoints found given the supplied search terms");
    }

    # save if they indicated we should
    $searchset->save_set($self->{'dbh'}, $options->{'name'})
      if ($options->{'name'} && $self->{'dbh'});

    return $searchset;
}

#######################################################################
# SCRIPTS                                                             #
#######################################################################

sub pull_gui_parts {
    my ($self, $buf) = @_;
    my @parts;
    foreach my $part (qw(title introduction)) {
	my ($extracted) = ($buf =~ /\%${part}\s*([^\%]+)/);
	$extracted =~ s/^\s*//;
	$extracted =~ s/\s*$//;
	push @parts, $extracted;
    }
    return @parts;
}

our $globalgui;

sub process_includes {
    my ($self, $buf) = @_;
    while ($buf =~ s/%include (.+)\n/$self->file_to_buf($1)/e) {
    }
    return $buf;
}

sub run_script {
    my ($self, $buf, $args) = @_;

    my $set = new GeoDB::Set;
    my $x;

    my $options_save = $self->{'options'};
    $self->{'options'} = undef;
    $self->parse_args($args);

    # convert from a file if needed
    $buf = $self->file_to_buf($buf)
      if ($buf =~ s/^file://);

    $buf = $self->process_includes($buf);

    # pull the title and introduction out
    my ($title, $intro) = $self->pull_gui_parts($buf);

    # process %[] params
    $buf =  $self->process_string_params($buf, '[', ']',
					 ($self->get_option('nogui') ? 0 : 1),
					 $title, $intro);
    ($buf) =  $self->deifify_text($buf);

    $globalgui = $self->get_option('nogui');
    $self->set_option('nogui', 1);

    GEODEBUG(1,"running a script\n");
    my $mode = '';
    my $text = '';
    my $scriptargs;
    my @scriptargs;
    my %scriptargs;

    while ($buf =~ s/^([^\n]*)\n//) {
	my $line = $1;
	next if ($line =~ /^\s*#/);
	next if ($line =~ /^\s*$/);
	if ($line =~ /^\%\s*(\w+)\s*(.*)/) {
	    my $newmode = $1;
	    my $newscriptargs = $2;
	    $self->run_mode_on_text($mode, $text, $set, $self->{'dbh'}, $args,
				    \%scriptargs);
	    $text = '';
	    $mode = $newmode;
	    $scriptargs = $newscriptargs;
	    @scriptargs = split(/ /,$scriptargs);
	    %scriptargs = map { split(/=/,$_); } @scriptargs;
	    next;
	} else {
	    if (!$scriptargs{'leavewhitespace'}) {
		$line =~ s/^\s*//;
		$line =~ s/\s*$//;
	    } else {
		$line .= "\n";
	    }
	    $text .= $line;
	}
    }
    $self->run_mode_on_text($mode, $text, $set, $self->{'dbh'}, $args,
			    \%scriptargs);
    $self->{'options'} = $options_save;
    $self->set_option('nogui', $globalgui);
}

#
# currently parses text only for %if/%else/%endif pairs.
#  - really not the right place to do this; should operate as working
#  - that would require a 'noop' mode for gui var extration...
sub deifify_text {
    my ($self, $buf) = @_;
    my $returntext = "";
    my ($addtext, $matchtext, $ignoretext);

    while ($buf =~ s/^([^\n]*\n)//) {
	my $line = $1;
	next if ($line =~ /^\s*#/);
	next if ($line =~ /^\s*$/);

	# look for marking section ( ^%KEYWORD )
	if ($line =~ /^\%\s*(endif|else)/) {
	    return ($returntext, $buf, $1);
	} elsif ($line =~ /^\%\s*(\w+)\s*(.*)/) {
	    my $keyword = $1;
	    my $otherargs = $2;
	    if ($keyword eq 'if') {
		if ($otherargs && eval "$otherargs") {
		    # true if statement
		    ($addtext, $buf, $matchtext) =
		      $self->deifify_text($buf);
		    $returntext .= $addtext;
		    if ($matchtext eq 'else') {
			($ignoretext, $buf, $ignoretext) =
			  $self->deifify_text($buf);
		    }
		} else {
		    # false if statement
		    ($ignoretext, $buf, $matchtext) =
		      $self->deifify_text($buf);
		    if ($matchtext eq 'else') {
			($addtext, $buf, $matchtext) =
			  $self->deifify_text($buf);
			$returntext .= $addtext;
		    }
		}
	    } else {
		$returntext .= $line;
	    }
	} else {
	    $returntext .= $line;
	}
    }
    return ($returntext, $buf, '');
}

sub run_mode_on_text {
    my ($self, $mode, $text, $set, $dbh, $args, $scriptargs) = @_;

    $dbh = $self->{'dbh'} if (!$dbh);
    if ($scriptargs->{'allowgui'}) {
	$self->set_option('nogui', $globalgui);
    }
    if ($mode eq 'search') {
	GEODEBUG(2,"script: searching using $text\n");
	$set->clear(); # XXX: make optional
	$self->search($text, { searchset => $set,
			       dbh => $dbh });
	GEODEBUG(3, "search results: " . $set->num() . "\n");
    } elsif ($mode eq 'export') {
	GEODEBUG(2,"script: exporting using $text\n");
	$self->export_it($text, $set);
    } elsif ($mode eq 'import') {
	GEODEBUG(2,"script: importing using $text\n");
	$self->import_it($text);
    } elsif ($mode eq 'modify') {
	GEODEBUG(2,"script: modifying using $text\n");
	$self->modify($set, $text);
    } elsif ($mode eq 'display') {
	GEODEBUG(2,"script: displaying using $text\n");
	$self->display($set, $text);
    } elsif ($mode eq 'print') {
	$text = $self->process_string_params($text,'[',']');
	GEODEBUG(2,"script: printing text\n");
	print $text,"\n";
    } elsif ($mode eq 'perl') {
	$text = $self->process_string_params($text,'[',']')
	  unless ($scriptargs->{'dontexpand'});
	GEODEBUG(2,"evaluating embedded perl\n");
	my $result = eval $text;
	if ($@) {
	    print STDERR "dump:$text\n";
	    GEOERR("evaluated perl died: $@\n");
	}
    } elsif ($mode eq 'script') {
	my ($scriptname, $args) = ($text =~ /([^:]+):*(.*)/);
	GEODEBUG(2,"script: running script $scriptname\n");
	$self->run_script("file:" . $scriptname, $args);
    } elsif ($mode eq 'ignore' || $mode eq 'introduction' || $mode eq 'title') {
	GEODEBUG(2,"Ignoring text in the ignore section\n");
    } elsif ($mode eq 'save') {
	GEODEBUG(2,"saving set back to the DB");
	$set->load();  # ensure everything got loaded before we save it!
	               # XXX: have the save routines only save what got loaded
	$set->save();
    } elsif ($mode eq '') {
	GEOERR("no mode to run script text on: $text\n") if ($text ne '');
    } else {
	GEOERR("unknown script mode: $mode\n");
    }
    $self->set_option('nogui', 1);
}


sub file_to_buf {
    my ($self, $file, $dirname) = @_;
    $dirname ||= "scripts";
    my $buf;
    my @searchpaths = ("", $self->get_config_dir() . "/$dirname/",
		      "/usr/share/geoqo/scripts/");
    foreach my $path (@searchpaths) {
	if (-f "$path$file") {
	    $file = "$path$file";
	    last;
	}
    }
    if (! -f $file) {
	if (-f $self->get_config_dir() . "/$dirname/" . $file) {
	    $file = $self->get_config_dir() . "/$dirname/" . $file;
	} else {
	    GEOERR("can't find script: $file\n");
	}
	
    }
    open(INFILE, $file);
    while (<INFILE>) {
	$buf .= $_;
    }
    close(INFILE);
    return $buf;
}

#
# loads display engines and calls them on a search set
#
sub display {
    my ($self, $searchset, $displist, $dbh) = @_;

    my @displist = split(/;/,$displist);

    foreach my $disparg (@displist) {
	my $engine = $self->new_engine($disparg, "Display");
	my $text = $engine->display($searchset, $dbh);
	if ($text) {
	    # don't create a new one if DNE already
	    my $qw = $self->{'qwizard'};
	    if ($qw && !$self->get_option('toconsole')) {
		# display results in the existing window
		my %primaries =
		  ('geoqodisplay' =>
		    { title => 'GeoQO Output',
		    name => 'geoqodisplay',
		    questions => 
		    [{ type => 'paragraph',
		       preformatted => 1,
		       values => sub { return [$text]; }
		     }]
		    });
		$qw->merge_primaries(\%primaries);
		$qw->magic('geoqodisplay')
	    } else {
		# display results on the console, or whatever.
		print $text if ($text);
	    }
	}
    }
}


######################################################################
# GUI
#

sub get_qwizard {
    my ($self, $brandnew) = @_;
    eval "require QWizard;" || return;
    require QWizard;
    import QWizard;
    $self = $self->get_master();
    return $self->{'qwizard'}
      if (exists($self->{'qwizard'}) && !$brandnew);

    $self->{'qwizard'} = new QWizard(title => 'GeoQO');
    return $self->{'qwizard'};
}

######################################################################
# PARSING
#
# parses a format of type:file or file.type and returns type,file
#
sub get_type_and_file {
    my ($f) = (@_);
    GEODEBUG(2,"examining file name $f\n");
    my ($type, $file, $args);

    # first determine if there is a prefix:REST syntax
    ($type, $file) = ($f =~ /([a-zA-Z0-9]+):(.*)/);
    if ($file) {
	# yay!
	# but... there may be arguments, ie: prefix:arg1=val1:file.something
	# in which case, arg1 and val1 may contain :s as well...
	# note that:
        #   system:cmd=gpsbable -i geo -f %{file:loc} -o ...
        # is common and thus you can't assume a :file.xyz suffix
	# thus the legal file parts are a-z,A-Z,0-9,. and /
	$f = $file;
	my $legal_file_pieces = '[-_\.a-zA-Z0-9\/\\]+';
	($args, $file) = ($f =~ /(.*):([-_\.a-zA-Z0-9\/\\]+)$/);
	if ($file) {
	    # we found both arguments and a file path
	    # (nothing to do)
	} else {
	    # no actual sub-pieces
	    if ($f =~ /^[-_\.a-zA-Z0-9\/\\]+$/) {
		# assume just a file name is present
		$file = $f;
	    } else {
		# A colon exists in it; assume not a file
		# just $type should be returned (like system: above)
		$args = $f;
		$file = "";
	    }
	}
    } else {
	# there is no type: prefix, so assume the entire thing is a path
	$file = $f;
	GEODEBUG(3,"determining file type for $file\n");
	($type) = ($file =~ /.*\.([a-zA-Z0-9]+)$/);
	GEODEBUG(4,"extension says $type\n");
	$type ||= 'gpx';	# default to gpx if nothing else
    }

    # now change the type to 1 upper case letter and everything else lower.
    $type =~ s/^(.)(.*)/uc($1) . lc($2)/e;

    $type ||= "";
    $args ||= "";
    $file ||= "";
    GEODEBUG(3,"file type parsing results: $type, $args, $file\n");
    return ("$type:$args", $file) if ($args);
    return ($type, $file);
}

#
# parse name=value,name2=value2 specifications and returns a hash
#
#   Also sets optionsarray and options in the (maybe) passed self ref
sub parse_args {
    my $self;
    $self = shift if (ref($_[0]) ne '');
    my ($specs, $args, %parse_options) = @_;

    $args = {} if (!defined($args));
    my @argarr;

    # retrieve the list of OPTION specs from the child module (if it has any) 
    my @options = eval "@" . ref($self) . "::GEOOPTIONS;";
    my $numopts = $#options;
    my %info = eval "%" . ref($self) . "::GEOINFO;";

    # specs = arg=val,...
    my @components = split(/,/, $specs);
    foreach my $c (@components) {
        if ($c =~ /([^=<>]+)(==|=-|=\+|=!|=|<=>|<>|<|>)(.*)/) {
	    $args->{$1} = $3;
	    push @argarr, [$1, $3, $2];
	} else {
	    # assume a pre-defined option spec
	    my $opt = shift @options;
	    GEOERR("unknown value passed to " . ref($self) . ": $c\n")
	      if (!$opt && $numopts > -1 && !$info{'allowextraargs'});
	    next if (!$opt);
	    $args->{$opt->{'name'}} = $c;
	    push @argarr, [$opt->{'name'}, $c, '='];
	}
    }

    # set the defaults for the remainder of unspecified options.  maybe.
    foreach my $option (@options) {
	if (!defined($args->{$option->{'name'}})) {
	    if (exists($option->{'isconfig'})) {
		$args->{$option->{'name'}} =
		  $self->get_config($option->{'name'})
		    if ($self->get_config($option->{'name'}));
		print "here: \n";
	    } else {
		$args->{$option->{'name'}} =
		  $self->get_default($option->{'name'})
		    if ($self->get_default($option->{'name'}));
	    }
	    $args->{$option->{'name'}} ||= $option->{'default'};
	}
    }

    $self->{'options'} = $args if ($self);
    $self->{'optionsarray'} = \@argarr if ($self);
    return \@argarr if ($parse_options{'wantarrayref'});
    return $args;
}

# replaces strings with %{name} tokens with option values of the same name
sub process_string_params {
    my ($self, $string, $left, $right, $usegui, $title, $introduction) = @_;
    $left ||= "{";
    $right ||= "}";
    my $regexp = "\%\\$left([^\\$right\/]+)/?([^\\/\\$right]*)/?([^\\$right]*)\\$right";
    my $qw;
    $qw = $self->get_qwizard()
      if ($usegui && $#{$self->{'optionsarray'}} == -1);

    if ($qw) {
	my @qs;
	my @options;
	my %doneit;
	while ($string =~ /$regexp/g) {
	    # format: name/default value/remaining stuff

	    my ($name, $default, $remaining) = ($1, $2, $3);
	    $name =~ s/\n*$//;
	    $default =~ s/\n*$//;
	    $remaining =~ s/\n*$//;

	    next if (exists($doneit{$name}));
	    $doneit{$name} = 1;
	    push @options, $name;

	    my %guiopts;
	    $guiopts{'default'} = $default;
	    $guiopts{'name'} = $name;
	    $guiopts{'text'} = $name;

	    # split the remaining pieces by /s
	    my @remainingparts = split(/\/\n*/,$remaining);
	    foreach my $part (@remainingparts) {
		my ($leftpart, $rightpart) = ($part =~ /([^=]+)=(.*)/);
		next if (!$rightpart);
		$guiopts{$leftpart} = $rightpart;
	    }

	    if (exists($guiopts{'isconfig'})) {
		$guiopts{'default'} =
		  $self->get_option($name) if ($self->get_option($name));
	    } else {
		$guiopts{'default'} =
		  $self->get_default($name) if ($self->get_default($name));
	    }

	    my $q = $self->create_question_from_args(\%guiopts);
	    push @qs, $q;
	}
	$qw->merge_primaries(
			     {
			      'scriptparams' => 
			      { title => ($title || 'Script Parameters'),
				introduction => $introduction,
				questions => \@qs,
				post_answers => 
				[[sub {
				      my $opts = $_[$#_-1];
				      my $self = $_[$#_];
				      foreach my $opt (@options) {
					  $self->{'options'}{$opt} = 
					    qwparam($opt)
					      if (qwparam($opt) ne '');
				      }
				  },\@options, $self]],
			      },
			     });
	$qw->magic('scriptparams') if ($#qs > -1);
    }
    $string =~ s/$regexp/$self->{'options'}{$1} || $2/eg;
    return $string;
}

#######################################################################
# PROGRESS BAR support
#
my $have_progress;
sub setup_progress {
    if (!$_[0]->{'no-progress'} && !$_[0]->get_option('no-progress') &&
	$_[0]->get_option('debug') && $_[0]->get_option('debug') < 4) {
	# don't event attempt to load this if no-progress is set so
	# that long load errors aren't printed to the console (happens
	# in make test)
	$have_progress = eval "require Term::ProgressBar";
	if ($have_progress) {
	    my ($self, $count, $title) = @_;
	    $self->{'progressbar'} =
	      Term::ProgressBar->new({count => $count || 1,
				      name => $title,
				      remove => 1,
				      ETA => 'linear'});
	    $self->{'progressbar'}->minor(0);
	    $self->{'progresscount'} = $count || 1;
	}
    }
}

sub set_progress {
    return 1000000 if ($_[0]->{'no-progress'} ||
		       $_[0]->get_option('no-progress'));
    if ($_[0]->{'progressbar'}) {
	return $_[0]->{'progressbar'}->update($_[1]);
    }
    return 1000000;  # if they're checking return codes for the next
                     # update count, this will likely keep them away for longer.
}

sub stop_progress {
    return if ($_[0]->{'no-progress'} || $_[0]->get_option('no-progress') || 
               !$_[0]->{'progressbar'});
    $_[0]->{'progressbar'}->update($_[0]->{'progresscount'});
}


######################################################################
# SUBCLASS HELPER FUNCTIONS
#

sub get_temp_dir {
    my $self = shift;
    return $self->{'options'}{'tmpdir'} || $self->get_option('tmpdir') ||
      $self->get_config_dir() . "/temp";
}

sub get_config_dir {
    my $self = shift;
    return $self->{'options'}{'configdir'} || $self->get_option('configdir') ||
      $ENV{'HOME'} . "/.geoqo";
}

1;
