package GeoDB::DB;

use strict;

use GeoDB::Utils;

my $DBVERSION = 9;

#
# creates a default database handle given the type of database we want
# to use, etc
#
sub init_dbh {
    my ($self, $type, $dbname, $host, $user, $pass, $dbport) = @_;
    my $dbh;
    return $self->{'dbh'} if ($self->{'dbh'});
    $type ||= 'sqlite2';

    if ($type eq 'mysql') {
	$dbh = DBI->connect("DBI:mysql:database=$dbname;host=$host",
			    $user, $pass);
    } elsif ($type eq 'sqlite2') {
	$dbh = DBI->connect("DBI:SQLite2:dbname=$dbname");
	$dbh->{'sqlite_handle_binary_nulls'}=1
    } elsif ($type eq 'sqlite') {
	$dbh = DBI->connect("DBI:SQLite:dbname=$dbname");
    } elsif ($type eq 'postgres') {
	my $portspec;
	$portspec = ";port=$dbport" if ($dbport);
	$dbh = 
	  DBI->connect("DBI:Pg:dbname=$dbname;host=$host$portspec",
		       $user, $pass);
    } else {
	return;
    }

    if (!$dbh) {
	print STDERR "failed to open connection to the geoqo database\n";
	exit(1);
    }
    $dbh->{'RaiseError'} = 1;
    init_sql_utils($dbh);
    $self->{'dbh'} = $dbh;

    $self->init_global($self->{'dbh'}) if ($self->{'dbh'});

    return $dbh;
}

#######################################################################
# DATABASE routines and definitions
#
my %DB_IDX = 
  (
   'mysql' => "int(11) NOT NULL auto_increment primary key",

   'sqlite2' => "integer primary key",

   'postgres' => 'serial primary key',

   'generic' => 'int',
  );


sub setup_db {
    my ($self, $type, $justone, $tmpname) = @_;
    my $dbh = $self->{'dbh'};
    my ($table, $idindex) = ("TABLE", ", primary key (ident)");

    if ($tmpname) {
	$table = "temporary TABLE";
	$idindex = "";
    }

    die("database type not specified for setup_db: type=$type") if (!$type);

    # create the storage tables
    if (!$justone || $justone eq 'waypoints') {
	GEODEBUG(2, "creating database table waypoints\n");
	$dbh->do("CREATE $table " . ($tmpname ? $tmpname : "waypoints") . " (
              ident                              varchar(255),
              descr                              varchar(4096),
              lat                                float,
              lon                                float,
              sym                                varchar(255),
              time                               timestamp,
              type                               varchar(255),
              subtype                            varchar(255),
              url                                varchar(4096),
              cmt                                varchar(4096),
              urlname                            varchar(255)

              $idindex
             )");
    }

    if (!$justone || $justone eq 'geocaches') {
	GEODEBUG(2, "creating database table geocaches\n");
	$dbh->do("CREATE $table " . ($tmpname ? $tmpname : "geocaches") . " (
              ident                              varchar(255)
                                      REFERENCES waypoints(ident),
              geocache_type                      varchar(255),
              groundspeak_id                     int,
              groundspeak_available              varchar(255),
              groundspeak_archived               varchar(255),
              groundspeak_name                   varchar(4096),
              groundspeak_placed_by              varchar(4096),
              groundspeak_owner_id               int,
              groundspeak_owner_name             varchar(4096),
              groundspeak_type                   varchar(4096),
              groundspeak_container              varchar(4096),
              groundspeak_difficulty             float,
              groundspeak_terrain                float,
              groundspeak_country                varchar(4096),
              groundspeak_state                  varchar(4096),
              groundspeak_short_description_html varchar(255),
              groundspeak_short_description_text varchar(4096),
              groundspeak_long_description_html  varchar(255),
              groundspeak_long_description_text  varchar(65536),
              groundspeak_encoded_hints          varchar(4096)

              $idindex
             )");
    }

    if (!$justone || $justone eq 'waymarks') {
	GEODEBUG(2, "creating database table waymarks\n");
	$dbh->do("CREATE $table " . ($tmpname ? $tmpname : "waymarks") . " (
              ident                              varchar(255)
                                      REFERENCES waypoints(ident),
              waymark_owner_name                 varchar(4096)

              $idindex
             )");
    }

    if (!$justone || $justone eq 'waypointdata') {
	GEODEBUG(2, "creating database table waypointdata\n");
	$dbh->do("CREATE $table " . ($tmpname ? $tmpname : "waypointdata") . " (
              ident      varchar(255)
                         REFERENCES waypoints(ident),
              attrname   varchar(1024),
              attrvalue  varchar(4096)
             )");
    }

    if (!$justone || $justone eq 'logs') {
	GEODEBUG(2, "creating database table logs\n");
	$dbh->do("CREATE $table " . ($tmpname ? $tmpname : "logs") . " (
              id           $DB_IDX{$type},
              ident        varchar(255)
                           REFERENCES waypoints(ident),
              date         timestamp,
              type         varchar(255),
              finder_id    int,
              finder       varchar(255),
              text_encoded varchar(255),
              text         varchar(4096)
            )");
    }

    if (!$justone || $justone eq 'sets') {
	GEODEBUG(2, "creating database table sets\n");
	$dbh->do("CREATE $table " . ($tmpname ? $tmpname : "sets") . " (
              name       varchar(255),
              ident      varchar(255)
            )");
    }

    if (!$justone || $justone eq 'parameters') {
	GEODEBUG(2, "creating table parameters\n");
	$dbh->do("CREATE $table " . ($tmpname ? $tmpname : "parameters") . " (
              name       varchar(255),
              value      varchar(4096)
            )");
	$dbh->do("INSERT into parameters(name, value) values('dbversion','" . $DBVERSION . "')");
    }

    if (!$justone || $justone eq 'dbversion') {
	$dbh->do("INSERT into parameters(name, value) values('dbversion','" . $DBVERSION . "')");
    }

    if (!$justone || $justone eq 'wigles') {
	GEODEBUG(2, "creating database table wigles\n");
	$dbh->do("CREATE $table " . ($tmpname ? $tmpname : "wigles") . " (
              ident                   varchar(255)
                                      REFERENCES waypoints(ident),
              ssid                    varchar(4096),
              netid                   varchar(4096),
              discover                varchar(4096),
              channel                 int,
              wigtype                 varchar(256),
              freenet                 varchar(16),
              firsttime               timestamp,
              wep                     varchar(16),
              comment                 varchar(4096),
              qos                     varchar(16),
              lastupdt                timestamp,
              paynet                  varchar(16),
              userfound               varchar(16)

              $idindex
             )");
    }

    if (!$justone || $justone eq 'log_ident_index') {
	GEODEBUG(2, "indexing log table on ident\n");
	$dbh->do("CREATE index log_ident_index on logs(ident)");
    }

    if (!$justone || $justone eq 'geodinings') {
	GEODEBUG(2, "creating database table geodinings\n");
	$dbh->do("CREATE $table " . ($tmpname ? $tmpname : "geodinings") . " (
              ident                   varchar(255)
                                      REFERENCES waypoints(ident),
              restaurant_name         varchar(256),
              submitted_by            varchar(256),
              price                   float,
              service                 float,
              continent               varchar(256),
              country                 varchar(256),
              state                   varchar(256),
              city                    varchar(256),
              zip                     int,
              address                 varchar(256),
              description             varchar(4096),
              details                 varchar(16)

              $idindex
             )");
    }

    if (!$justone || $justone eq 'tags') {
	GEODEBUG(2, "creating database table tags\n");
	$dbh->do("CREATE $table " . ($tmpname ? $tmpname : "tags") . " (
              ident                   varchar(255)
                                      REFERENCES waypoints(ident),
              tagname                 varchar(255),
              ismine                  int,
              needspublished          int,
              count                   int,
              datefetched             timestamp,
              datepublished           timestamp
             )");
	$dbh->do("CREATE index tag_ident_index on tags(ident)");
	$dbh->do("CREATE index tag_tagname_index on tags(tagname)");
    }

    if (!$justone || $justone eq 'ratings') {
	GEODEBUG(2, "creating database table ratings\n");
	$dbh->do("CREATE $table " . ($tmpname ? $tmpname : "ratings") . " (
              ident                   varchar(255)
                                      REFERENCES waypoints(ident),
              ratingname              varchar(255),
              ratingvalue             float,
              othersrating            float,
              needspublished          int,
              count                   int,
              datefetched             timestamp,
              datepublished           timestamp
             )");
	$dbh->do("CREATE index rate_ident_index on ratings(ident)");
	$dbh->do("CREATE index rate_ratingname_index on ratings(ratingname)");
    }

}

sub update_one {
    my ($self, $type, $name, $vertocheck, $num) = @_;

    if ($$num == $vertocheck) {
	GEODEBUG(3,"upgrading DB from version $vertocheck -> " . 
		 ($vertocheck+1) . "\n");
	$self->setup_db($type, $name);
	$$num += 1;
    }
}

sub update_db {
    my ($self, $type) = @_;
    my $dbh = $self->{'dbh'};
    GEOERR("update_db called illegally ($dbh)\n") if (!$dbh);

    $dbh->{'RaiseError'} = 0;
    my $sth = $dbh->prepare_cached("select value from parameters where name = 'dbversion'");
    $sth->execute();

    my $ver = '';
    if ($sth && !$sth->errstr) {
	# no version
	$ver = $sth->fetchrow_arrayref();
	$ver = $ver->[0] if (ref($ver) eq 'ARRAY');
    }
    $dbh->{'RaiseError'} = 1;

    $sth->finish(); # required to work around a bug.

    GEODEBUG(2,"DB upgrade: current version = $ver, software version = $DBVERSION\n");
    if ($ver eq $DBVERSION) {
	GEODEBUG(3,"DB version matches current schema; not updating\n");
	return;
    }

    if ($ver && $ver > $DBVERSION) {
	GEOERR("Your database was created using a newer version of geoqo/geodb; bailing out!");
    }

    $dbh->begin_work();

    # upgrade from 0 or nothing to 1
    if (!$ver || $ver == 0) {

	GEODEBUG(3,"upgrading 0 -> 1\n");

	if ($ver eq '') {
	    $self->setup_db($type, "parameters");
	    $dbh->do("insert into parameters(name, value) values('dbversion','0')");
	}

	$self->setup_db($type, "waypoints", "tmpwaypoints");
	$dbh->do("insert into tmpwaypoints(ident, descr, lat, lon, sym, time, type, subtype, url, urlname) select ident, descr, lat, lon, sym, time, type, subtype, url, urlname from waypoints");
	$dbh->do("drop table waypoints");
	$self->setup_db($type, "waypoints");
	$dbh->do("insert into waypoints(ident, descr, lat, lon, sym, time, type, subtype, url, urlname) select ident, descr, lat, lon, sym, time, type, subtype, url, urlname from tmpwaypoints");

	$ver++;
    }

    if ($ver == 1) {
	GEODEBUG(3,"upgrading 1 -> 2\n");
	$dbh->do("drop table geocachedata");
	$self->setup_db($type, "waypointdata");
	$ver++;
    }

    if ($ver == 2) {
	GEODEBUG(3,"upgrading 2 -> 3\n");
	$dbh->do("drop table waypointdata");
	$self->setup_db($type, "waypointdata");
	$ver++;
    }

    $self->update_one($type, "waymarks", 3, \$ver);
    $self->update_one($type, "wigles", 4, \$ver);
    $self->update_one($type, "log_ident_index", 5, \$ver);
    $self->update_one($type, "geodinings", 6, \$ver);
    $self->update_one($type, "tags", 7, \$ver);
    $self->update_one($type, "ratings", 8, \$ver);

    $dbh->commit();

    # odd issues with multiple rows and use of update; this is safer.
    $dbh->do("delete from parameters where name = 'dbversion'");
    $dbh->do("insert into parameters(name, value) values('dbversion',$ver)");

    GEODEBUG(1,"upgraded DB to version $ver\n");
}

1;
