package GeoDB::Display::Top10;

use strict;
use GeoDB::Display;
use GeoDB::Utils;

our @ISA = qw(GeoDB::Display);

our $VERSION = "0.71";

our @GEOOPTIONS = 
  (
   { name => 'field',
     help => 'The waypoint field name to count.',
     default => 'groundspeak_owner_name'},
   { name => 'show',
     help => 'The number of top things to show.',
     default => 10 },
   { name => 'doaverage',
     title => 'Display the average value:',
     helpdesc => '(Only useful for numeric fields)',
     type => 'checkbox',
   },
   { name => 'averagetext',
     title => 'Average Text Prefix:',
     default => 'Average:',
   },
   { name => 'script',
     title => 'Script name to run on each result:',
     helpesc => '%s will be replaced by the value',
   },
   { name => 'noblank',
     title => 'Do not show empty valued results',
     default => 1,
     type => 'checkbox'
   },

  );

sub display {
     my ($self, $set, $dbh) = @_;

     my $field = $self->{'options'}{'field'};
     GEODEBUG(3, "Counting placers\n");
     $set->load({nologs => 1});
     my %counts;
     my $count = 0;
     my $total = 0;
     my $out;
     my $shownum = $self->{'options'}{'show'};
     if ($shownum > 0) {
	 $set->foreach(sub {
			   $counts{$_[0]{'data'}{$field}}++;
			   $count++;
		       } );
     }
     if ($self->{'options'}{'doaverage'}) {
	 $set->foreach(sub {
			   $total += $_[0]{'data'}{$field};
		       });
     }

     my @keys = keys(%counts);
     @keys = sort {$counts{$b} <=> $counts{$a}} @keys;

     my $setnum = $set->num();
     if ($setnum == 0) {
	 $out = "No results in set; can't calculate top 10 results\n";
	 return $out;
     }
     if ($shownum) {
	 $out .= "Top $shownum '${field}'s\n";
	 $out .= sprintf("%4s %4s %s\n",'Num','%','Value');
	 $out .= "-" x 70 ."\n";
	 for (my $i = 0 ; $i < $shownum && $i <= $#keys; $i++) {
	     next if ($self->{'options'}{'noblank'} && $keys[$i] eq '');
	     $out .= sprintf("%4d %4d %s\n",$counts{$keys[$i]}, int(100*$counts{$keys[$i]}/$setnum), $keys[$i]);
	 }
     }
     if ($self->{'options'}{'doaverage'}) {
	 $out .= sprintf("%s %0.5f\n", $self->{'options'}{'averagetext'},
			 ($total/$count));
     }
     if ($self->{'options'}{'script'}) {
	 my $script = $self->{'options'}{'script'};
	 my ($scriptname, $args) = ($script =~ /([^:]+):*(.*)/);
	 for (my $i = 0 ; $i < $shownum && $i <= $#keys; $i++) {
	     next if ($self->{'options'}{'noblank'} && $keys[$i] eq '');
	     GEODEBUG(2,"script: running script $scriptname on $keys[$i]\n");
	     my $master = $self->get_master();
	     $master->run_script("file:" . $scriptname,
				 sprintf($args, $keys[$i]));
	 }
     }
     return $out;
 }


1;


=head1 NAME

top10 - Analyze the data set to count fields of similar types 

=head1 DESCRIPTION

START HELP

The top10 dispaly plugin counts the values in a particular field and
shows you the top10 results of the counts in that field.

Example (defaults to counting the top 10 waypoint owners):

  > geoqo -s any: -d top10
  Top 10 'groundspeak_owner_name's
   Num    % Value
  ----------------------------------------------------------------------
  22    4 BootyBuddies
  20    4 sailorscodgers
  17    3 retiredprof
  16    3 two bison
  14    2 Sac D
  14    2 Oldhippy & Granny
  13    2 TRAKD
  12    2 rivercity
  11    2 2Dee2Dee
  10    2 jimlips
  10    2 Pirate Princesses

Example: counting container types

  > geoqo -s any: -d top10:groundspeak_container
  Top 10 'groundspeak_container's
   Num    % Value
  ----------------------------------------------------------------------
   234   50 Micro
   103   22 Regular
    92   19 Small
    13    2 Not chosen
    12    2 Other
    10    2 Virtual
     4    0 Large

Setting the "doaverage" parameter will also display the average value
associated with a numeric field:

  > geoqo -s 'cache:owner_name=Yamar' -d top10:groundspeak_difficulty,doaverage=1
  Top 10 'groundspeak_difficulty's
   Num    % Value
  ----------------------------------------------------------------------
    14   38 2
     9   25 1.5
     6   16 1
     3    8 2.5
     2    5 3
     1    2 4
     1    2 3.5
  Average: 1.90278

END HELP

=cut
