package GeoDB::Export::Kml;

use strict;
use GeoDB::Export;
use GeoDB::Utils;
use GeoDB::Export::Density;
use IO::File;

our $VERSION = "0.71";

our @ISA = qw(GeoDB::Export GeoDB::Export::Density);

our @GEOOPTIONS =
  ({ name => 'size',
     default => 200,
     title => 'Size of density plot',
     help => 'Creates a plot where there are SIZE by SIZE squares of colored rectangles'
   },
   { name => 'width',
     indent => 1,
     title => 'Width of density plot',
     help => 'Normally equal to the size value but can be set independently.  It specifies the width (in squares) of the density plot.'
   },
   { name => 'height',
     indent => 1,
     title => 'Height of density plot',
     help => 'Normally equal to the size value but can be set independently.  It specifies the height (in squares) of the density plot.'
   },
   { name => 'spread',
     title => 'Spreading Distance',
     help => 'Sets the spread that determines the number of neighboring squares that a given cache will effect.  EG, a spread of 5 will affect a circle of density squares in a radius of 5.  This generally should be a low value of probably not more than 5 or so.',
     default => 5,
   },
   { name => 'altitude',
     title => 'Altitude of density plots',
     help => 'The altitude to set the density plot squares at.',
     default => 3000
   },
   { name => 'extrude',
     title => 'Extrude the density plot to the ground:',
     help => 'Whether or not to extrude the density plot squares down to the ground.',
     type => 'checkbox',
     default => 0,
   },
   { name => 'doempty',
     title => 'Include empty squares in the density plot:',
     help => 'If set to 1, even empty squares will be colored (red).  Otherwise the empty squares are removed from the plot leaving the plane ground underneath.',
     type => 'checkbox',
     default => 0,
   },
   { name => 'opaque',
     title => 'Opaqueness of the density squares:',
     helpdesc => '(0-100; 0 being completely see-through)',
     help => 'The opacity of the density squares.  Must be between 0 and 100, with 0 being completely see-through (and thus completely useless).',
     default => 20,
   },
   { name => 'linewidth',
     title => 'Line width:',
     help => 'The width of the border lines to draw (0-4).  0 means don\'t draw borders on the squares.',
     default => 0,
   },
   { name => 'nmax',
     title => 'Maximum North coordinate:',
     help => 'The maximum north coordinate value to created the grid over.  Normally this is automatically set by the maximum value found in the data, but can be overridden',
   },
   { name => 'nmin',
     title => 'Minimum North coordinate:',
     help => 'The minimum north coordinate value to created the grid over.  Normally this is automatically set by the minimum value found in the data, but can be overridden'
   },
   { name => 'wmax',
     title => 'Maximum West coordinate:',
     help => 'The maximum west coordinate value to created the grid over.  Normally this is automatically set by the maximum value found in the data, but can be overridden'
   },
   { name => 'wmin',
     title => 'Minimum West coordinate:',
     help => 'The minimum west coordinate value to created the grid over.  Normally this is automatically set by the minimum value found in the data, but can be overridden'
   }
  );

sub _x {
    return GeoDB::Export::_add_basic_entity(@_);
}

sub export_it {
    my ($self, $file, $set) = @_;

    GEODEBUG(1,"writing file: $file\n");
    my $fh = new IO::File ">$file";

    # start the header
    # <bounds minlat="38.236517" minlon="-122.04275" maxlat="38.870533" maxlon="-121.271167" />
    print $fh "<?xml version=\"1.0\" encoding=\"utf-8\"?>
<kml xmlns=\"http://earth.google.com/kml/2.0\">
<Folder>
  <description>GeoQO Data Exports</description>
";

    $set->load({nologs => 1});

    if (!$self->{'options'}{'nopoints'}) {

print $fh "  <Folder>
    <name>GeoQO Data Export 1</name>\n";
    $set->foreach(sub {
		      my $c=$_[0]{'data'};
		      my $type = ($c->{'subtype'} ? "$c->{'type'}|$c->{'subtype'}" : $c->{'type'});
		      print $fh "
  <Placemark>
    <description>" . _x($c->{'urlname'}) . "</description>
    <name>" . _x($c->{'ident'}) . "</name>
    <styleUrl>#khStyle652</styleUrl>
    <Point>
      <altitudeMode>clampToGround</altitudeMode>
      <coordinates>" . _x($c->{'lon'}) . "," . _x($c->{'lat'}) . ",0</coordinates>
    </Point>
  </Placemark>
";
		  });


    print $fh "</Folder>\n";
    }

    $self->get_density_map($set, $fh) if (!$self->{'options'}{'nodensity'});

print $fh "</Folder>
</kml>
";
}

1;

=cut

=head1 NAME

Kml - export kml file for Google Earth

=head1 DESCRIPTION

START HELP

  > geoqo -s 'set:import:near me' -e fortheearth.kml

The software will also produce a density plot as well, which shows the
density of caches in particular areas color-coded by relief colors
(rainbow; red = very few, purple = many).  Examples of this can be
seen on Yamar's cache homepage page http://geocaching.sharedlists.com/.

Options:

=over

=item size=NUM

  > geoqo -s 'set:import:near me' -e kml:size=100:fortheearth.kml

Changes the number of squares in the density plot to NUM by NUM.
Defaults to 50.

=item width=NUM

=item height=NUM

Changes just the width and height of the squares.  The size parameter
can change both at once.

=back

(there are a bunch of options for the density plot, which are
currently undocumented).

END HELP

=head1 AUTHOR

Wes Hardaker (AKA Yamar) <hardaker ATAT users.sourceforge.net> 

=cut
