package GeoDB::Import::Gpx;

require GeoDB::Import;
use GeoDB::Utils;
use GeoDB::Set;
use XML::Simple;
use strict;

our $VERSION = "0.71";

our @ISA = qw(GeoDB::Import);
our @GEOOPTIONS = 
({ name => 'setname',
   help => 'Uses setname as the name of geoqo SET to create based on the data. Normally this is extracted from the I<name> field within the gpx file itself and prepended with "import:".  IE, if you had a pocket query named "my search" then, when you imported the gpx file into geoqo, you\'d get a new set named "import:my search".'
 });

my @existingtags = qw(lat lon time sym desc type url urlname);
my %existingtags;
map { $existingtags{$_} = 1 } @existingtags;

sub import_set {
    my ($self, $file, $dbh, %opts) = @_;
    GEODEBUG(1,"reading file: $file\n");
    $XML::Simple::PREFERRED_PARSER = "XML::Parser";
    my $xdoc = XMLin($file);

    my $set = new GeoDB::Set;
    # get the name
    my $name = $self->{'options'}{'setname'} || ('import:' . $xdoc->{'name'});
    GEODEBUG(2,"processing file: $file into set $name\n");

    # read in each waypoint (cache)
    my $count = 0;

    # XXX: check for an array instead, which would indicate no name
    # field was present in the gpx for the waypoint...  should probably
    # handle this or at least error.
    if (ref($xdoc->{'wpt'}) eq 'ARRAY') {
	GEOERR("unhandled GPX format; please contact geoqo developers\n");
    }

    # protect against singular records
    if (exists($xdoc->{'wpt'}{'desc'})) {
	$xdoc = { wpt => { $xdoc->{'wpt'}{'name'} => $xdoc->{'wpt'} } };
    }

    if (exists($xdoc->{'creator'}) && $xdoc->{'creator'} eq 'Geodining.com') {
	my ($gd, $type, $subtype, $waypoint);
	while (my ($ident, $wpt) = each(%{$xdoc->{'wpt'}})) {
	    GEODEBUG(3, "importing geodining #" . $count++ . ": $ident\n");

	    # these names need to be carefully matched.  we're translating
	    # names in the Geocache.pm definition to the names in the XML,
	    # which are almost equal in most tags except for a few differences.
	    $gd = $wpt->{'geodining:restaurant'};
	    ($type, $subtype) = ('Geodining', $wpt->{'type'});
	    $waypoint =
	      new GeoDB::Waypoint
		({
		  data => 
		  {
		   ident =>                      "$ident",
		   'desc' => 	     	         "$wpt->{'desc'}",
		   'lat' => 	     		 "$wpt->{'lat'}",
		   'lon' => 	     		 "$wpt->{'lon'}",
		   'sym' => 	     		 "$wpt->{'sym'}",
		   'time' =>       		 "$wpt->{'time'}",
		   'type' => 	     	         "$type",
		   'subtype' => 	     	 "$subtype",
		   'url' => 	     		 "$wpt->{'url'}",
		   'urlname' =>     		 "$wpt->{'urlname'}",
		   'id' => 	     		 "$gd->{'id'}",

		   'geodining_restaurant_name' =>
		   "$gd->{'geodining:restaurant_name'}",

		   'geodining_submitted_by' =>	 
		   "$gd->{'geodining:submitted_by'}",

		   'geodining_price' =>		 "$gd->{'geodining:price'}",
		   'geodining_service' =>	 "$gd->{'geodining:service'}",
		   'geodining_continent' =>	 "$gd->{'geodining:continent'}",
		   'geodining_country' =>   	 "$gd->{'geodining:country'}",
		   'geodining_state' =>   	 "$gd->{'geodining:state'}",
		   'geodining_zip' =>   	 "$gd->{'geodining:zip'}",
		   'geodining_address' =>   	 "$gd->{'geodining:address'}",
		   'geodining_description' =>    "$gd->{'geodining:description'}",
		   'geodining_details' =>   	 "$gd->{'geodining:details'}",
		  }});
	    $waypoint->morph_to_specific();
	    $set->add($waypoint);
	}
	$name = $main::opts->{'name'} || $name;
	$set->set_name($name) if ($name);
	return $set;
    }


    # assume geocaching
    # import each cache in the series
    my ($gs, $waypoint, $type, $subtype);
    while (my ($ident, $wpt) = each(%{$xdoc->{'wpt'}})) {
	GEODEBUG(3, "importing cache #" . $count++ . ": $ident\n");

	# these names need to be carefully matched.  we're translating
	# names in the Geocache.pm definition to the names in the XML,
	# which are almost equal in most tags except for a few differences.
	$gs = $wpt->{'groundspeak:cache'};
	($type, $subtype) = ($wpt->{'type'} =~ /(.*)\|(.*)/);
	($type, $subtype) = ($wpt->{'type'}, '') if (!$type);
	$waypoint =
	  new GeoDB::Waypoint
	    ({
	     data => 
	     {
	      ident =>                         "$ident",
	      'desc' => 	     	       "$wpt->{'desc'}",
	      'lat' => 	     		       "$wpt->{'lat'}",
	      'lon' => 	     		       "$wpt->{'lon'}",
	      'sym' => 	     		       "$wpt->{'sym'}",
	      'time' =>       		       "$wpt->{'time'}",
	      'type' => 	     	       "$type",
	      'subtype' => 	     	       "$subtype",
	      'url' => 	     		       "$wpt->{'url'}",
	      'urlname' =>     		       "$wpt->{'urlname'}",
	      'id' => 	     		       "$gs->{'id'}",
	      'available' =>  		       "$gs->{'available'}",
	      'archived' =>   		       "$gs->{'archived'}",
	      'groundspeak_name' => 	       "$gs->{'groundspeak:name'}",
	      'groundspeak_placed_by' =>       "$gs->{'groundspeak:placed_by'}",
	      'groundspeak_owner_id' =>
	      "$gs->{'groundspeak:owner'}{'id'}",

	      'groundspeak_owner_name' =>
	      "$gs->{'groundspeak:owner'}{'content'}",
	      'groundspeak_type' => 	       "$gs->{'groundspeak:type'}",
	      'groundspeak_container' =>       "$gs->{'groundspeak:container'}",
	      'groundspeak_difficulty' =>      "$gs->{'groundspeak:difficulty'}",
	      'groundspeak_terrain' => 	       "$gs->{'groundspeak:terrain'}",

	      'groundspeak_country' => 	       "$gs->{'groundspeak:country'}",
	      'groundspeak_state' => 	       "$gs->{'groundspeak:state'}",

	      'groundspeak_short_description_html' =>
	      "$gs->{'groundspeak:short_description'}{'html'}",
	      'groundspeak_short_description' =>
	      "$gs->{'groundspeak:short_description'}{'content'}",

	      'groundspeak_long_description_html' =>
	      "$gs->{'groundspeak:long_description'}{'html'}",
	      'groundspeak_long_description' =>
	      substr("$gs->{'groundspeak:long_description'}{'content'}",0,65535),

	      'groundspeak_encoded_hints' =>
	      ((ref($gs->{'groundspeak:encoded_hints'}) eq 'HASH') ? "" : "$gs->{'groundspeak:encoded_hints'}"),
	     }});
	$waypoint->morph_to_specific();
	my $logs = $gs->{'groundspeak:logs'}{'groundspeak:log'};
	if (exists($logs->{'id'})) {
	    # protect against singular logs
	    $logs = {$logs->{'id'} => $logs };
	}
	while (my ($logid, $log) = each(%$logs)) {
	    GEODEBUG(4, "importing log #" . $logid . "\n");
	    $waypoint->{'logs'}{$logid} = 
	      {
	       'id' => "$logid",
	       'date' => "$log->{'groundspeak:date'}",
	       'type' => "$log->{'groundspeak:type'}",
	       'finder_id' => "$log->{'groundspeak:finder'}{'id'}",
	       'finder' => "$log->{'groundspeak:finder'}{'content'}",
	       'text' => "$log->{'groundspeak:text'}{'content'}",
	       'text_encoded' => "$log->{'groundspeak:text'}{'encoded'}"
	      };
	}
	$set->add($waypoint);
    }
    $name = $main::opts->{'name'} || $name;
    $set->set_name($name) if ($name);
    return $set;
}

1;

=cut

=head1 NAME

Gpx - import gpx files

=head1 DESCRIPTION

START HELP

  > geoqo -i myfile.gpx

  > geoqo -i gpx:setname=myname:myfile.gpx

The gpx importer imports gpx files taken from the geocaching.com site
(and probably many other sites too, but only geocaching.com GPX files
have been tested so for) and imports the data into your geoqo database.

END HELP

=head3 Options:

=over

=item setname=NAME

Uses setname as the name of geoqo SET to create based on the data.
Normally this is extracted from the I<name> field within the gpx file
itself and prepended with "import:".  IE, if you had a pocket query
named "my search" then, when you imported the gpx file into geoqo, you'd
get a new set named "import:my search".

=back

=head1 AUTHOR

Wes Hardaker (AKA Yamar) <hardaker ATAT users.sourceforge.net> 

=cut
