package GeoDB::Import::Zip;
use Cwd;

require GeoDB::Import;

use GeoDB;
use GeoDB::Utils;
use GeoDB::Set;
use GeoDB::Import::Gpx;
use strict;

maybe_load('Archive::Zip', "You need the Archive::Zip perl module to support loading of zip files\n", 'qw(:ERROR_CODES)');

our $VERSION = "0.71";

our @ISA = qw(GeoDB::Import);

our @GEOOPTIONS = 
(
 { name => 'setname',
   help => 'Uses setname as the name of geoqo SET to create based on the data. Normally this is extracted from the I<name> field within the gpx file itself and prepended with "import:".  IE, if you had a pocket query named "my search" then, when you imported the gpx file into geoqo, you\'d get a new set named "import:my search".'},
 { name => 'tmpdir',
   help => 'The zip importer works by copying the zip file to a tmpdir and then unzipping it there.  This is normally set to I<$HOME/.geoqo/tmpdir>. This option lets you override that default.'
 },
);

sub import_set {
    my ($self, $file, $dbh, %opts) = @_;
    GEODEBUG(1,"unzipping file: $file\n");

    my $tmpdir = $self->{'options'}{'tmpdir'} || $ENV{'HOME'} . "/.geoqo/temp";
    mkdir($tmpdir) if (! -d $tmpdir);    my $olddir = getcwd();


    my $zip = Archive::Zip->new();
    my $zipstatus = $zip->read($file);

    die "Unable to open $file!\n" if $zipstatus != 'AZ_OK';

    GEODEBUG(5,"File Opened: $file\n");

    # Ok - we opened it, lets make a set to return
#    chdir($tmpdir);
    my $set = new GeoDB::Set();
    $set->set_name(($self->{options}{setname}) ? $self->{options}{setname} :
                    "import:" . $file); 
    # We only care about file types we handle.

    my %wptsetnames;
    foreach my $zipmember ($zip->members()) {
      my $zipf = $zipmember->fileName();
      GEODEBUG(5,"Considering $zipf\n");
      my ($type, $file) = GeoDB::get_type_and_file($zipf);
      my $engine = $self->new_engine("$type:nogui=1", 'Import');
      if ($engine && ref($engine) =~ /GeoDB::Import/) {
	GEODEBUG(5,"... $type - Importing\n");
	$zip->extractMemberWithoutPaths($zipmember, $tmpdir . "/" . $zipf);

	# geocaching.com smart setname hack (see below)
	$engine->{'options'}{'setname'} = $wptsetnames{$zipf}
	  if (exists($wptsetnames{$zipf}));
	
	# pass in our set name
	$engine->{'options'}{'setname'} = $self->{'options'}{'setname'}
	  if ($self->{'options'}{'setname'});

	my $tmpset = $engine->import_it($tmpdir . "/" . $zipf, $dbh);


	# geocaching.com smart setname hack:
	# setname was specified.  remember a variation of this
	# file name to best support geocaching.com extra waypoint
	# files which end in -wpts.
	if ($tmpset->get_name() =~ /import:/) {
	    $zipf =~ s/\.gpx/-wpts.gpx/;
	    $wptsetnames{$zipf} = $tmpset->get_name() . ":waypoints";
	}

        GEODEBUG(5,$tmpset->num() . " waypoints loaded.\n");
	$set->merge($tmpset);
        GEODEBUG(5,$set->num() . " waypoints in merged set.\n");
	unlink($tmpdir . "/" . $zipf);
      } else {
	GEODEBUG(5,"...Skipping\n");
      }
    }

#    chdir($olddir);
    GEODEBUG(5,$set->num() . " waypoints in set we return.\n");
    return $set;

  }

1;

=cut

=head1 NAME

Zip - import zipped files

=head1 DESCRIPTION

START HELP

  > geoqo -i myfile.zip

The I<zip> importer allows you to simply specify a zip file for
importing instead of having to unzip it yourself.  It expects the
contents of the zip file to be named similar to the zip file itself,
just like groundspeak sends them out for geocaching.com pocket
queries.  IE, if you import "111111.zip" it'll try to unzip it and
then import the contained 111111.gpx file.  It'll also try to import
the 1111111-wpts.gpx file, if one existed as well.

Note: using this importer requires you to have B<unzip> installed on
your system.

END HELP

=head3 Options:

=over

=item tmpdir=DIR

The zip importer works by copying the zip file to a tmpdir and then
unzipping it there.  This is normally set to I<$HOME/.geoqo/tmpdir>.
This option lets you override that default.

=back

END HELP

=head1 AUTHOR

Wes Hardaker (AKA Yamar) <hardaker ATAT users.sourceforge.net> 

=cut
