#!/usr/bin/perl

package GeoDB::Modify::Classify;

use strict;
use GeoDB::Modify;
use GeoDB::Utils;
use AI::Categorizer;
use AI::Categorizer::Collection::Files;
use AI::Categorizer::Collection::InMemory;
use File::Spec;


# ai::cat requires:
#     Params::Validate
#     Lingua::Stem
#     Algorithm::NaiveBayes
#     Class::Container
#     Statistics::Contingency

# Lingua::Stem requires:
#     Lingua::PT::Stemmer
#     Lingua::Stem::It
#     Lingua::Stem::Snowball::No
#     Text::German
#     Lingua::Stem::Snowball::Se
#     Lingua::Stem::Fr
#     Lingua::GL::Stemmer
#     Lingua::Stem::Snowball::Da
#     Lingua::Stem::Ru

# AI::Categorizer::Learner::DecisionTree requires:
#     AI::DecisionTree

# AI::Categorizer::Learner::SVM requires:
#     Algorithm::SVM


our @ISA = qw(GeoDB::Modify);

our $VERSION = "0.71";

our @GEOOPTIONS =
({ name => 'inputprefix',
   title => 'Input Prefix',
   default => '',
 },
 {
  name => 'outputprefix',
  title => 'Output Prefix',
  default => 'predicted:'},
 {
  name => 'categories',
  title => 'Category List:',
  default => 'kids,coolcontainer,goodlocation,fun'
 },
 {
  name => 'LearningType',
  title => 'Learning Method',
  default => 'DecisionTree',
  values => [qw(DecisionTree NaiveBayes SVM)],
  type => 'menu',
 },
 {
  name => 'stopwordsfile',
  title => 'Stop-words file:',
  default => '/usr/share/geoqo/classify-ignore-words'
 },
 {
  name => 'Debugging',
  title => 'Print Verbose Debugging Output:',
  type => 'checkbox',
 },
);

sub load_a_cache {
    my ($cache, $self, $trainingset, $outset) = @_;
    my $data = $cache->{'data'};
    my $tags = $cache->{'tags'};

    my $point = 
      {
       name => $data->{'ident'},
       stemming => 'porter',
       content_weights =>
       { groundspeak_short_description => 2,
	 groundspeak_long_description => 2,
	 urlname => 4,
	 groundspeak_owner_name => 2,
       },
      };
    foreach my $token (qw(urlname
			  groundspeak_owner_name
			  groundspeak_long_description
			  groundspeak_short_description
			  groundspeak_difficulty
			  groundspeak_terrain
			  groundspeak_type
			  groundspeak_hints)) {
	$point->{'content'}{$token} =
	  $data->{$token} || "";
	$point->{'content_weights'}{$token} ||= 1;
    }

    my $prefix = $self->{'options'}{'inputprefix'};
    @{$point->{'categories'}} =
      grep {exists($tags->{$prefix . $_}) && $tags->{$prefix . $_}}
	@{$self->{'categories'}};

    my @marked =
      grep {exists($tags->{$prefix . $_})}
	@{$self->{'categories'}};

    # XXX: do logs

    if ($#marked == -1) {
	# not catagories, this one is to analyize later
	$outset->{$data->{'ident'}} = $point;
    } else {
	$trainingset->{$data->{'ident'}} = $point;
    }
}


sub modify_set {
    my ($self, $set) = @_;

    my (%trainingset, %outset);

    my $corpus = $self->{'options'}{'stopwordsfile'};
    my $stopwords = File::Spec->catfile( $corpus, 'stopwords' );
    my $v = $self->{'options'}{'Debugging'};

    # load the data and store in required format by the categorizer
    GEODEBUG(2,"Loading data for the catgeroizer\n");
    $set->load();

    GEODEBUG(3,"Parsing data for the catgeroizer\n");
    @{$self->{'categories'}} = split(/,\s*/, $self->{'options'}{'categories'});
    $set->foreach(\&load_a_cache, $self,
		  \%trainingset, \%outset);

    # start the training based on existing data
    my $training =
      AI::Categorizer::Collection::InMemory->new(data => \%trainingset,
						 stopword_file => $stopwords);
    my $k = AI::Categorizer::KnowledgeSet->new( verbose => $v );

    GEODEBUG(3,"Loading the data into the knowledge set\n");
    $k->load( collection => $training );

    ######################################## TRAINING
    GEODEBUG(2,"training the categorizer\n");

    my $l;
    if ($self->{'options'}{'LearningType'} eq 'DecisionTree') {
	maybe_load('AI::Categorizer::Learner::DecisionTree',
		  'The AI::Categorizer::Learner::DecisionTree perl module is required for using DecisionTrees to classify caches');
	$l =  AI::Categorizer::Learner::DecisionTree->new( verbose => $v );
    } elsif ($self->{'options'}{'LearningType'} eq 'NaiveBayes') {
	maybe_load('AI::Categorizer::Learner::NaiveBayes',
		  'The AI::Categorizer::Learner::NaiveBayes perl module is required for using NaiveBayes to classify caches');
	$l =  AI::Categorizer::Learner::NaiveBayes->new( verbose => $v );
    } elsif ($self->{'options'}{'LearningType'} eq 'SVM') {
	maybe_load('AI::Categorizer::Learner::SVM',
		  'The AI::Categorizer::Learner::SVM perl module is required for using SVM to classify caches');
	$l =  AI::Categorizer::Learner::SVM->new( verbose => $v );
    } else {
	GEOERR("unknown type of learning method: " .
	       $self->{'options'}{'LearningType'});
    }
    $l->train( knowledge_set => $k );

    ######################################## ANALYZING
    my $outdata =
      AI::Categorizer::Collection::InMemory->new(data => \%outset,
						 stopword_file => $stopwords);

    my $prefix = $self->{'options'}{'outputprefix'};
    while (my $document = $outdata->next) {
	my $result = $l->categorize( $document );
	my @categories = $result->categories;
	my $cache = $set->find_by_ident($document->{'name'});
	map { $cache->set_tag($prefix . $_) } @categories;
    }
}

1;

