package GeoDB::Search;

=cut

=head1 NAME

GeoDB::Search - module structure for searching a GeoDB database

=head1 SYNOPSIS

This module is merely a parent class for real search modules.

=head1 DESCRIPTION

BEGIN HELP

The searching features within geoqo let you search for waypoints in
various ways.  See the documentation for the sub-modules for futher
details.

The list of search modules distributed with GeoQO are:

=for comment
START LIST

  Cache
  Waymark
  Geodining
  Wigle
  Waypoint
  Ident
  Log
  Attr
  Rating
  Tag
  Set
  Any

=for comment
END LIST

To see help on any of the above SUBTOPICs, run:

  > geoqo -d help:topic=search/SUBTOPIC

END HELP

When the GeoDB->search() function (geoqo -s STRING from the command
shell) is called it will parse strings passed to it into modules.
These modules are automatically loaded from the perl path and
GeoDB::Search parent hierarchy.  Thus, a search string of
"NAME:param1=foo,param2=bar" will cause GeodB to autoload
GeoDB::Search::Name (note the uppercase on only the first letter
here!).  Then the search function on that loaded module will be called
with two arguments: a pointer to itself and the search string to parse
("param1=foo,param2=bar" in the above example).  The search function
should then return a portion of a sql where clause that will be
inserted into a greater statement.  It should also return an array
reference to the arguments to fill that clause.

As an example, if param1=foo,param2=bar was passed as the $list (second) argument then the search function might:

  return ("something = ? && somethingelse = ?", ['foo', 'bar'])

The new function is typically inherited and doesn't need to be
created.

To write a real search module use the following template replacing
NAME by your search name:

  package GeoDB::Search::NAME;

  use strict;
  require GeoDB::Search;
  our $VERSION = "0.71";
  our @ISA = qw(GeoDB::Search);

  sub search {
      my ($self, $list) = @_,;
      my $clause, @args;

      GEODEBUG(3, "searching using NAME: $list\n");

      # parse and deal with $list and then return a sql where clause string

      return ($clause, \@args);
  }

  1;

=head1 EXAMPLES

Look at the GeoDB::Search::Cache module for an example.

=cut

use strict;
use GeoDB::Utils;

our @ISA = qw(GeoDB);

sub get_search_subexpression {
    my ($term, $eq, $value, $valarray) = @_;
    if ($term !~ /\?/) {
	$term = "$term ARG ?";
    }
    if ($eq eq '=') {
	$value =~ s/\*/%/g;
	$value = "%" . $value . "%";  # force fuzziness
	$eq = 'like';
    } elsif ($eq eq '<>') {
	$value =~ s/\*/%/g;
	$value = "%" . $value . "%";  # force fuzziness
	$eq = 'not like';
    } elsif ($eq eq '==') {
	# exact match no auto-fuzzy
	if ($value =~ s/\*/%/g) {
	    $eq = 'like';
	} else {
	    $eq = "=";
	}
    } elsif ($eq eq '<=>') {
	# exact match no auto-fuzzy
	if ($value =~ s/\*/%/g) {
	    $eq = 'not like';
	} else {
	    $eq = "<>";
	}
    }
    $term =~ s/ ARG / $eq /g;
    push @$valarray, $value;
    return $term;
}

sub get_search_tokens {
    my ($self, $criteria, $type, $cols, $colmap, $startclause) = @_;

    GEODEBUG(3, "breaking down criteria: $criteria\n");
    $criteria = $self->parse_args($criteria, undef, wantarrayref => 1);

    # build a where clause; not exactly "secure" though
#    my @bases = split(/,/, $criteria);  # split by ,s into indiv criteira

#     my @criteria;
#     map {
# 	my @parts = split(/=/,$_); # split a criteria by the =
# 	GEOERROR("unable to parse criteria $criteria\n") if ($#$bases > 1);
# 	@parts = ('.', $_) if ($#parts == 0);
# 	$parts[1] = "%" . $parts[1] . '%';  # always add fuzziness XXX: optional
# 	push @criteria, \@parts;
#     } @bases;

    # allow mulitple hits...
    my $wherestr;
    $wherestr = "$startclause (" if ($startclause);
    my @whereargs;
    foreach my $crit (@$criteria) {
	my $matched = 0;
	foreach my $col (@$cols) {
	    if ($crit->[0] eq $col) {
		# exact match, use it
		$wherestr .= "and (" .
		  get_search_subexpression($colmap->{$col} || $col, $crit->[2],
					   $crit->[1], \@whereargs) . ")";
		$matched = 1;
		last;
	    }
	}
	next if ($matched);

	$wherestr .= " and (";
	foreach my $col (@$cols) {
	    if ($col =~ /^$crit->[0]/) {
		# beginning match, use it
		$wherestr .= " or (" .
		  get_search_subexpression($colmap->{$col} || $col, $crit->[2],
					   $crit->[1], \@whereargs) . ")";
		$matched = 1;
		# allow for other matches too... so don't bail
	    } elsif ($col =~ /_$crit->[0]/) {
		# suffix match after a :, use it
		$wherestr .= " or (" .
		  get_search_subexpression($colmap->{$col} || $col, $crit->[2],
					   $crit->[1], \@whereargs) . ")";
		$matched = 1;
		# allow for other matches too... so don't bail
	    }
	}
	$wherestr .= ")";
	$wherestr =~ s/\( or /\(/;
	if (!$matched) {
	    GEOERR("Can't find a field name match for criteria: '$crit->[0]' of type '$type'\n");
	}
    }

    GEODEBUG(3,"out: $wherestr\n");
    GEODEBUG(3,"arg count: $#whereargs\n");
    $wherestr =~ s/\( *and /\(/;
    $wherestr =~ s/^ *and //;
    $wherestr .= ")" if ($startclause);

    return ($wherestr, \@whereargs);
}

1;
