# ==============================================================================
# Alfandega::Config.pm
# ------------------------------------------------------------------------------
# Copyright(c) 2002 Christian Tosta <c_tosta@yahoo.com.br>. All rights reserved.
# This program is free software; you can redistribute it and/or modify it under
# the same terms as Perl itself.
#
# =+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+
#
#                            Firewall Config Parser
#
# =+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

package Alfandega::Config;

require 5.004;

use strict;
use vars qw(@ISA @EXPORT $VERSION);
use Alfandega::Language;
use Net::IP;

@ISA = qw(Exporter);
@EXPORT = qw(locale);
$VERSION = "2.1";

sub new {
    my $class = shift;
    my $file  = shift;
    my $option;
    my %Config;
    my %ACLs;
    my %Default = (
        'accept_redirects'                     =>   'off',
        'accept_source_route'                  =>   'off',
        'anywhere'                             =>   '0/0',
        'any-port'                             =>   '0:',
        'config_basedir'                       =>   '/etc/alfandega',
        'acl_config'                           =>   '/etc/alfandega/acl.conf',
        'conntrack_ftp_support'                =>   'on',
        'conntrack_irc_support'                =>   'off',
        'default_forward_policy'               =>   'DROP',
        'default_input_policy'                 =>   'DROP',
        'default_nat_prerouting_policy'        =>   'DROP',
        'default_nat_postrouting_policy'       =>   'DROP',
        'default_nat_output_policy'            =>   'DROP',
        'default_limit_burst'                  =>   '3',
        'default_limit_rate'                   =>   '3/minute',
        'default_log_level'                    =>   'info',
        'default_output_policy'                =>   'DROP',
        'distro_name'                          =>   'redhat',
        'firewall_manager'                     =>   'root',
        'icmp_echo_ignore_all'                 =>   'off',
        'icmp_echo_ignore_broadcasts'          =>   'on',
        'icmp_ignore_bogus_error_responses'    =>   'off',
        'ip_dynaddr'                           =>   'off',
        'ip_forward'                           =>   'on',
        'iptables_dump_file'                   =>   '/etc/sysconfig/iptables',
        'language'                             =>   'English',
        'load_kernel_modules'                  =>   'off',
        'log_default_policy'                   =>   'on',
        'log_martians'                         =>   'on',
        'minimal_config_version'               =>   '2.0',
        'nat_ftp_support'                      =>   'on',
        'nat_irc_support'                      =>   'off',
        'overwrite_sysctl'                     =>   'on',
        'priv_ports'                           =>   '1:1023',
        'remote_blacklist_method'              =>   'http_tranfer',
        'remote_blacklist_server'              =>   'http://alfandega.sourceforge.net/blacklist/',
        'rp_filter'                            =>   'on',
        'server_type'                          =>   'dual',
        'show_conntrack_info'                  =>   'on',
        'tcp_syncookies'                       =>   'on',
        'unpriv_ports'                         =>   '1024:65535',
        'Xwindow_ports'                        =>   '6000:6039',
    );
    my %Required = (
        'internal_interface'                   =>   'eth0',
        'external_interface'                   =>   'eth1',
    );
    
    foreach $option (keys %Default) {
        $Config{$option} = $Default{$option};
    }

    die "$file: ",$! if ! -e $file;
    open(CFG,'<',$file);
    while(<CFG>) {
        study;
        my ($option,$value) = s/^\s*(.*?)\s*$/$1/ &&
        /^(\S+)\s+(\S+)\s*(.*)/ if !/^#/;
        $Config{$option} = $value if $option;
    }
    close(CFG);

    chdir $Config{'config_basedir'} || die locale("Config::chdir_basedir");
    
    chmod 0755,"interfaces.conf", "scripts/adsl-auth"; # To review

    my @files;

    do {
        opendir(DIR, '.') || die locale("Config::open_dir"),$!;
        while (my $file = readdir DIR) {
            next if $file =~ /^\./;
            next if ! -x $file;
            push @files,$file if -f $file;
        }
        sleep 300 unless @files;
    } until @files;

    foreach my $file (@files) {
        next if $file =~ "README";
        open(CFG,'<',$file);
        while(<CFG>) {
            study;
            my ($option,$value) = s/^\s*(.*?)\s*$/$1/ &&
            /^(\S+)\s+(\S+)\s*(.*)/ if !/^#/;
            $Config{$option} = $value if $option;
        }
        close(CFG);
    }

    Alfandega::Language->new($Config{'language'});

    foreach $option (keys %Required) {
        die locale("Config::not_specified") . "'$option'\n" if !$Config{$option};
    }

    if ($Config{'config_version'} < $Config{'minimal_config_version'}) {
        die locale ("Config::bad_config_version");
    }

    open(ACLCFG,'<',$Config{'acl_config'}) or die locale("Config::acl_file"),$!;
    while(<ACLCFG>) {
        study;
        s/^\s*(.*?)\s*$/$1/;
        my ($version) = /^config_version\s+(\S+)\s*(.*)/;

        if (defined $version && $version < $Config{'minimal_config_version'}) {
            die locale ("Config::bad_acl_config_version");
        }

        my ($acl,$type,$regex) = /^acl\s+(\S+)\s+(\S+)\s+(\S+)\s*(.*)/ if !/^#/;

        next if !defined $acl;
        die locale("Config::acl_syntax_error"),$acl if !defined $type || $type =~ "";
        die locale("Config::acl_syntax_error"),$acl if !defined $regex || $regex =~ "";

        if($type =~ "ip_regex") {
            open(REGEXCFG,'<',$regex);
            my $i = 0;
            while(<REGEXCFG>) {
                study;
                s/^\s*(.*?)\s*$/$1/;
                my $pattern = substr($acl,0,-1);
                my ($ipstr) = /^$pattern\s+(\S+)\s*(.*)/ if !/^#/;
                next if !$ipstr;
                my $ip = new Net::IP("$ipstr") or die(Net::IP::Error());
                $Config{$acl} .= $ip->ip() . ",";

            }
            close(REGEXCFG);
            $Config{$acl} = substr($Config{$acl},0,-1) if $Config{$acl};
        }

        if($type =~ "ports") {
            $Config{$acl} = $regex;
        }

        if($type =~ "exec") {
            $Config{$acl} = $regex;
        }
    }
    close(ACLCFG);

    return %Config;
}

__END__
